/*

Copyright (C) 2000 Matthew P. Hodges
This file is part of XMakemol.

XMakemol is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

XMakemol is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with XMakemol; see the file COPYING.  If not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifdef GL

#define __GL_FUNCS_C__

#include <stdio.h>
#include <math.h>
#include <Xm/Xm.h>

#include "bonds.h"
#include "defs.h"
#include "draw.h"
#include "globals.h"
#include "view.h"

#include "gl_funcs.h"

/* Function prototypes */

double get_atom_scale (void);
double get_bond_width (void);
double get_hbond_scale (void);
double gl_get_rendered_atom_size (int);
void gl_init (void);
void gl_render_atoms (void);
void gl_render_bonds (void);
void gl_render_hbonds (void);
void gl_render_vectors (void);
void gl_set_perspective (void);
void gl_update_near_and_far (void);
double mod_of_vec (double *);
Boolean mouse_motion_p (void);
void normalize_vec (double *);
Boolean outline_mode_p (void);

/* Variable declarations */

static int gl_initialized = 0;
static int gl_no_atom_segments = 16;
static int gl_no_bond_segments = 12;
static int gl_use_lighting = 1;
static int gl_use_specular_lighting = 0;
static int gl_sep = 30; /* aro */
static GLfloat gl_shininess[] = {64};

static double gl_eye = 10;         /* per-system setting */
static double gl_fov = 50;
static double gl_max_z, gl_min_z;
static double gl_perspective_near = 0, gl_perspective_far = 0;

static GLubyte measure_stipple[] = {
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
  0xAA, 0xAA, 0xAA, 0xAA, 0x00, 0x00, 0x00, 0x00,
};

static GLUquadricObj *gl_quadric = NULL;

/* enum types */

enum gl_render_types
{
  BALL_AND_STICK,
  TUBES
};

static enum gl_render_types gl_render_type = BALL_AND_STICK;

void gl_render (void)
{
  enum render_stereo_types get_gl_render_stereo();
  void set_gl_render_stereo(enum render_stereo_types);
  void StereoProjection(float left, float right, float bottom,
                        float top, float near, float far,
                        float zero_plane, float dist, float eye);
  
  double eye_offset, gl_fov_rad, zp_width; /* zero parallax width */
  static double last_eye_offset;

  if (gl_initialized == 0)
    {
      gl_init ();
      gl_initialized = 1;
    }

  gl_set_perspective ();

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();

  gluLookAt (0.0, 0.0, gl_eye, 0.0, 0.0, -1.0, 0.0, 1.0, 0.0);

  /* aro--> */
  if(get_gl_render_stereo())
    {
      /* Calculate offset between eyes.  Max distance between eyes should be 
         (distance to projection plane) / 15 */
      eye_offset = (gl_eye + gl_perspective_near) / gl_sep;
      if(eye_offset > 0.0)
        last_eye_offset = eye_offset;
      else
        eye_offset = last_eye_offset;
      
      /* Calculate width at zero parallax (default z = 0) from given fov. 
         zp_width = (1/2) entire width at zero parallax */
      gl_fov_rad = (PI * gl_fov) / 180;
      zp_width = gl_eye * tan( 0.5 * gl_fov_rad );
      
      /* Left eye: */

      /* Create nonsymetric viewing frustum */
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      StereoProjection(-zp_width, zp_width, -zp_width, zp_width, 
                       gl_perspective_near, gl_perspective_far, 
                       0.0, gl_eye, -eye_offset);

      if(get_gl_render_stereo() == RED_BLUE) {
        glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glColorMask(GL_TRUE, GL_FALSE, GL_FALSE, GL_TRUE);/* Red mask */
      }
      else
        fprintf(stderr, "gl_render: Invalid stereo mode\n");

      glMatrixMode (GL_MODELVIEW);
      glLoadIdentity ();
    
      if ((! mouse_motion_p ()) || (! outline_mode_p ()))
        {

          if (gl_use_lighting)
            {
              glEnable (GL_LIGHTING);
              gluQuadricNormals (gl_quadric, GLU_SMOOTH);
            }
          else
            {
              glDisable (GL_LIGHTING);
              gluQuadricNormals (gl_quadric, GLU_NONE);
            }
          
          glEnable (GL_DEPTH_TEST);
          glEnable (GL_COLOR_MATERIAL);
          
          gluQuadricDrawStyle (gl_quadric, GLU_FILL);
          
          gl_render_atoms ();
          gl_render_bonds ();
          gl_render_hbonds ();
          gl_render_vectors ();
        }
      else
        {

          glDisable (GL_LIGHTING);
          glDisable (GL_DEPTH_TEST);
          glDisable (GL_COLOR_MATERIAL);

          gluQuadricDrawStyle (gl_quadric, GLU_LINE);
          gluQuadricNormals (gl_quadric, GLU_NONE);

          gl_render_atoms ();
        }

      /* Right eye: */

      /* Create nonsymetric viewing frustum */
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      StereoProjection(-zp_width, zp_width, -zp_width, zp_width, 
                       gl_perspective_near, gl_perspective_far, 
                       0.0, gl_eye, eye_offset);
      
      if(get_gl_render_stereo() == RED_BLUE) {
        glColorMask(GL_FALSE, GL_FALSE, GL_TRUE, GL_TRUE);/* Blue mask */
        glClear(GL_DEPTH_BUFFER_BIT);
      }
      else
        fprintf(stderr, "gl_render: Invalid stereo mode\n");
                
      glMatrixMode (GL_MODELVIEW);
      glLoadIdentity ();
    }
  else /* If not stereo, we need to clear color and depth */
    {
      glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
    }
  /* <--aro */

  if ((! mouse_motion_p ()) || (! outline_mode_p ()))
    {
      
      if (gl_use_lighting)
        {
          glEnable (GL_LIGHTING);
          gluQuadricNormals (gl_quadric, GLU_SMOOTH);
        }
      else
        {
          glDisable (GL_LIGHTING);
          gluQuadricNormals (gl_quadric, GLU_NONE);
        }

      glEnable (GL_DEPTH_TEST);
      glEnable (GL_COLOR_MATERIAL);

      gluQuadricDrawStyle (gl_quadric, GLU_FILL);

      gl_render_atoms ();
      gl_render_bonds ();
      gl_render_hbonds ();
      gl_render_vectors ();
    }
  else
    {

      glDisable (GL_LIGHTING);
      glDisable (GL_DEPTH_TEST);
      glDisable (GL_COLOR_MATERIAL);

      gluQuadricDrawStyle (gl_quadric, GLU_LINE);
      gluQuadricNormals (gl_quadric, GLU_NONE);
      
      gl_render_atoms ();
    }

  /* aro - change color mask so all of RGB color fields are cleared */
  if(get_gl_render_stereo() == RED_BLUE) 
    glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
  
  /* Note that we are drawing onto the canvas directly, since I
     haven't been able to figure out the cause of crashes (in
     glClearColor) on some displays. */

  glXSwapBuffers (display, XtWindow (canvas));

  /* Note that we should make this conditional; we only need it when
     XpmWriteFileFromPixmap is called and canvas_pm is used. */

  XCopyArea (display, XtWindow (canvas), canvas_pm, gc, 0, 0,
             canvas_width, canvas_width, 0, 0);

}

void StereoProjection(float left, float right, float bottom, float top,
                      float near, float far, float zero_plane, float dist,
                      float eye)
{
  enum render_stereo_types get_gl_render_stereo();
  
  float xmid,
    ymid,
    clip_near,
    clip_far,
    topw,
    bottomw,
    leftw,
    rightw,
    dx,
    dy,
    n_over_d;
  
  static double last_near;
  static double last_far;
  
  dx = right - left;
  dy = top - bottom;
  
  xmid = (right + left) / 2.0;
  ymid = (top + bottom) / 2.0;
  
  /* If the near clipping plane is going to be behind the eye, clamp its 
     position at that of the eye (We set it 1.0 first because glFrustum
     requires it to be non-zero */
  if(near > 0.0)
    { 
      last_near = clip_near = near;
      last_far = clip_far = far;
    }
  else
    {
      clip_near = last_near;
      clip_far = last_far;
    }
  
  n_over_d = clip_near / dist;
  
  topw = n_over_d * dy / 2.0;
  bottomw = -topw;
  rightw = n_over_d * (dx / 2.0 - eye);
  leftw  = n_over_d *(-dx / 2.0 - eye);
  
  /* Need to be in projection mode for this. */
  glLoadIdentity();
  glFrustum(leftw, rightw, bottomw, topw, clip_near, clip_far);
  
  glTranslatef(-xmid - eye,  -ymid,  -zero_plane - dist);
}

void gl_render_atoms (void)
{
  void gl_render_atom (int);

  int i;

  /* Return if atoms not being display and in ball and stick mode;
     note that we draw atoms if tubes mode if the bond is visible */

  if ((atom_flag == 0) && (gl_render_type == BALL_AND_STICK)) return;

  for (i = 0; i < no_atoms; i++)
    {

      if ((atoms[i].visi == 1) || (view_ghost_atoms == 1))
        {
          gl_render_atom (i);
        }
    }
}

void gl_render_atom (int i)
{
  enum render_stereo_types get_gl_render_stereo(void);/* aro */
  double atom_size;
  GLfloat red, green, blue;

  /* For tubes mode, we don't draw an atom (a) when bonds are not
     displayed and the atom is bonded and (b) when atoms are not
     displayed and the atom is not bonded. In essense, atoms and bonds
     are mutually exclusive. */

  if (gl_render_type == TUBES)
    {
      if (((bond_flag == 0) && (bond_adjacency_list[i] != NULL)) ||
          ((atom_flag == 0) && (bond_adjacency_list[i] == NULL))) return;
    }

  red   = atoms[i].red   / 65536.0;
  blue  = atoms[i].blue  / 65536.0;
  green = atoms[i].green / 65536.0;

  /* aro - We need to convert the colors to grayscale in order for 
     red and blue color masks to work properly */
  if(get_gl_render_stereo() == RED_BLUE)
    {
      red = (red + blue + green) / 3;
      blue = red;
      green = red;
    }

  glPushMatrix ();
  glTranslatef (atoms[i].x, atoms[i].y, atoms[i].z);

  /* If the atom is selected, draw a black disk behind the atom and
     stipple the atom itself */
   
  if(atoms[i].sel == 1)
    {

      glPolygonStipple (measure_stipple);
      glColor3f (0.0, 0.0, 0.0);

/*        gluDisk (quadric, */
/*                 0.0, */
/*                 atom_scale * atoms[i].cov_rad, */
/*                 gl_no_atom_segments, */
/*                 1); */

/*        gluSphere (quadric, */
/*                   atom_scale * atoms[i].cov_rad, */
/*                   gl_no_atom_segments, */
/*                   gl_no_atom_segments); */

      glEnable (GL_POLYGON_STIPPLE);
    }

  if ((! mouse_motion_p ()) || (! outline_mode_p ()))
    {
      glColor3f (red, green, blue);
    }
  else
    {
      glColor3f (0.0, 0.0, 0.0);
    }

  if (gl_render_type == BALL_AND_STICK)
    {
      if (bond_adjacency_list[i] == NULL)
        {
          /* If no bond, use van der Waals radius  */
          atom_size = get_atom_scale () * atoms[i].vdw_rad;
        }
      else
        {
          atom_size = get_atom_scale () * atoms[i].cov_rad;
        }
    }
  else if (gl_render_type == TUBES)
    {
      /* If we are not bonded to any other atoms, use normal size */

      if (bond_adjacency_list[i] == NULL)
        {
          atom_size = get_atom_scale () * atoms[i].vdw_rad;
        }
      else
        {
          /* We're getting the diameter, and we want the radius */
          atom_size = get_bond_width () / 2.0;
        }
    }
  else
    {
      atom_size = 0.0;
    }

  /* Set up rendering for elliptical atoms */

  if (((gl_render_type == BALL_AND_STICK) ||
       ((gl_render_type == TUBES) && bond_adjacency_list[i] == NULL)) &&
      (atoms[i].is_ellipse == 1))
    {
      void euler_to_matrix (double *, double *);

      int j, k;

      double axis[3], angle, matrix[9];

      double get_angle_axis (double *);

      /* Apply the rotation for the orientation of the system */

      for (j = 0; j < 3; j++)
        {
          for (k = 0; k < 3; k++)
            {
              angle_axis_matrix[j][k] = global_matrix[j][k];
            }
        }

      angle = get_angle_axis (axis);
      glRotatef (-angle / DEG2RAD, axis[0], axis[1], axis[2]);

      /* Apply the rotation describing the original orientation of
         the ellipse */

      euler_to_matrix (atoms[i].euler, matrix);

      for (j = 0; j < 3; j++)
        {
          for (k = 0; k < 3; k++)
            {
              angle_axis_matrix[j][k] = matrix[(j * 3) + k];
            }
        } 

      angle = get_angle_axis (axis);
      glRotatef (angle / DEG2RAD, axis[0], axis[1], axis[2]);

      /* Apply the anisotropic scaling */

      glScalef (atoms[i].shape[0], atoms[i].shape[1], atoms[i].shape[2]);
       
    }

  if ((! mouse_motion_p ()) || (! outline_mode_p ()))
    {
      gluSphere (gl_quadric,
                 atom_size,
                 gl_no_atom_segments,
                 gl_no_atom_segments);
    }
  else
    {
      gluDisk (gl_quadric,
               atom_size,       /* Inner radius */
               atom_size,       /* Outer radius */
               8,               /* Small number for slices */
               1);              /* Number of rings */
    }

  glPopMatrix ();

  glDisable (GL_POLYGON_STIPPLE);

}


void gl_render_bonds (void)
{
  void gl_render_bond (int, int);

  int i, j;
  struct node *ptr;

  if (bond_flag == 0) return;

  for (i = 0; i < no_atoms; i++)
    {      

      ptr = bond_adjacency_list[i];

      while (ptr != NULL)
        {
          j = (ptr->v);

          if (atoms[i].visi && atoms[j].visi)
            {
              gl_render_bond (i, j);
            }

          ptr = ptr->next;
        }
    }
}

void gl_render_bond (int i,
                     int j)
{
  enum render_stereo_types get_gl_render_stereo(void);/* aro */

  double angle, b[3], bond_radius, length, mod_b;
  GLfloat red, green, blue;

  /* We want the radius, ie half the width which represents the
     diameter */

  bond_radius = get_bond_width () / 2.0;

  /* b is the vector along the bond */

  b[0] = atoms[j].x - atoms[i].x;
  b[1] = atoms[j].y - atoms[i].y;
  b[2] = atoms[j].z - atoms[i].z;
          
  mod_b = mod_of_vec (b);
  normalize_vec (b);

  angle = acos (b[2]) / DEG2RAD;

  red   = atoms[i].red   / 65536.0;
  blue  = atoms[i].blue  / 65536.0;
  green = atoms[i].green / 65536.0;

  /* aro - We need to convert the colors to grayscale in order for 
     red and blue color masks to work properly */
  if(get_gl_render_stereo() == RED_BLUE)
    {
      red = (red + blue + green) / 3;
      blue = red;
      green = red;
    }
  
  glColorMaterial (GL_FRONT, GL_DIFFUSE);
  glColor3f (red, green, blue);

  glPushMatrix ();

  glTranslatef (atoms[i].x, atoms[i].y, atoms[i].z);

  glRotatef (180.0, 0.0, 0.0, 1.0);
  glRotatef (angle, b[1], -1.0 * b[0], 0.0);

  /* length of bond segment */

  length = (mod_b * atoms[i].cov_rad /
            (atoms[i].cov_rad + atoms[j].cov_rad));

  gluCylinder (gl_quadric,
               bond_radius,
               bond_radius,
               length,
               gl_no_bond_segments,
               1);

  glPopMatrix ();

}

void gl_render_hbonds (void)
{

  void gl_render_hbond (int, int);

  int i, j;
  struct node *ptr;

  if ((hbond_flag == 0) || (any_hydrogen == 0)) return;

  for (i = 0; i < no_atoms; i++)
    {      

      ptr = hbond_adjacency_list[i];

      while (ptr != NULL)
        {
          j = (ptr->v);

          if (i > j)           /* don't do things twice */
            {
              if (atoms[i].visi && atoms[j].visi)
                {
                  gl_render_hbond (i, j);
                }
            }

          ptr = ptr->next;
            
        }
    }
}


void gl_render_hbond (int i,
                      int j)
{
  int k, l, no_segments = 5;
  double angle, atom_size, bond_length, segment_length, hbond_width;
  double start[3], end[3], bond_vector[3], segment[3];

  /* The whole bond */

  bond_vector[0] = atoms[j].x - atoms[i].x;
  bond_vector[1] = atoms[j].y - atoms[i].y;
  bond_vector[2] = atoms[j].z - atoms[i].z;

  normalize_vec (bond_vector);

  /* The visible portion of the bond */

  atom_size = gl_get_rendered_atom_size (i);

  start[0] = atoms[i].x + (atom_size * bond_vector[0]);
  start[1] = atoms[i].y + (atom_size * bond_vector[1]);
  start[2] = atoms[i].z + (atom_size * bond_vector[2]);

  atom_size = gl_get_rendered_atom_size (j);

  end[0] = atoms[j].x - (atom_size * bond_vector[0]);
  end[1] = atoms[j].y - (atom_size * bond_vector[1]);
  end[2] = atoms[j].z - (atom_size * bond_vector[2]);

  for (k = 0; k < 3; k++)
    {
      bond_vector[k] = end[k] - start[k];
    }

  bond_length = mod_of_vec (bond_vector);
  segment_length = bond_length / ((2 * no_segments) - 1);

  normalize_vec (bond_vector);
  angle = acos (bond_vector[2]) / DEG2RAD;

  glColorMaterial (GL_FRONT, GL_DIFFUSE);
  /* aro - as long as hbonds are always grayscale, they'll
     render properly in red/blue stereo mode */
  glColor3f (0.0, 0.0, 0.0);

  for (k = 0; k < no_segments; k++)
    {
      /* Coordinate for start of segment */

      for (l = 0; l < 3; l++)
        {
          segment[l] = start[l] + (k * 2) * bond_vector[l] * segment_length;
        }

      glPushMatrix ();
      glTranslatef (segment[0], segment[1], segment[2]);

      glRotatef (180.0, 0.0, 0.0, 1.0);
      glRotatef (angle, bond_vector[1], -1.0 * bond_vector[0], 0.0);

      /* length of bond segment */

      hbond_width = 0.15 * get_hbond_scale ();

      gluCylinder (gl_quadric,
                   hbond_width,
                   hbond_width,
                   segment_length,
                   gl_no_bond_segments,
                   1);

      glPopMatrix ();

    }
}

double gl_get_rendered_atom_size (int i)
{
  double atom_size = 0;

  /* Work out the size the atom is going to be rendered at taking into
     account the drawing mode (BALL_AND_STICK or TUBES) and whether or
     not atoms and bonds are being displayed */

  if (gl_render_type == BALL_AND_STICK)
    {
      if (atom_flag)
        {
          /* If not bonded use VDW, otherwise covalent radius */

          if (bond_adjacency_list[i] == NULL)
            {
              atom_size = get_atom_scale () * atoms[i].vdw_rad;
            }
          else
            {
              atom_size = get_atom_scale () * atoms[i].cov_rad;
            }
        }
      else
        {
          atom_size = 0;        /* Atoms not being rendered */
        }
    }
  else if (gl_render_type == TUBES)
    {
      if (bond_adjacency_list[i] == NULL)
        {
          /* Atom is not bonded; only render if atoms being displayed */

          if (atom_flag)
            {
              atom_size = get_atom_scale () * atoms[i].vdw_rad;
            }
          else
            {
              atom_size = 0;
            }
        }
      else
        {
          /* Atom is bonded; only render if bonds are being displayed */

          if (bond_flag)
            {
              atom_size = get_bond_width () / 2;
            }
          else
            {
              atom_size = 0;
            }
        }
    }

  return (atom_size);

}

void gl_render_vectors (void)
{

  void gl_render_vector (int, int);

  int i, j;

  if (vector_flag == 0)
    {
      return;
    }

  glColorMaterial (GL_FRONT, GL_DIFFUSE);
  glColor3f (0.5, 0.5, 0.5);

  for (i = 0; i < no_atoms; i++)
    {
      for (j = 0; j < MAX_VECTORS_PER_ATOM; j++)
        {
          if (atoms[i].has_vector > j)
            {
              gl_render_vector (i, j);
            }
        }
    }
}

void gl_render_vector (int i, int j)
{

  double get_vector_arrow_angle (void);
  double get_vector_arrow_scale (void);
  double get_vector_scale (void);

  double angle, height, mod_of_vector, top_radius;
  double vector_arrow_angle, vector_arrow_scale, vector_scale;
  double vector[3];

  vector_arrow_angle = get_vector_arrow_angle ();
  vector_arrow_scale = get_vector_arrow_scale ();
  vector_scale = get_vector_scale ();
 
  /* Draw stem */

  vector[0] = atoms[i].v[j][0] * vector_scale;
  vector[1] = atoms[i].v[j][1] * vector_scale;
  vector[2] = atoms[i].v[j][2] * vector_scale;

  mod_of_vector = mod_of_vec (vector);

  normalize_vec (vector);

  height = mod_of_vector * vector_arrow_scale;

  angle = acos (vector[2]) / DEG2RAD;

  glPushMatrix ();

  glTranslatef (atoms[i].x, atoms[i].y, atoms[i].z);
  glRotatef (180.0, 0.0, 0.0, 1.0);
  glRotatef (angle, vector[1], -1.0 * vector[0], 0.0);

  gluCylinder (gl_quadric,
               0.025,
               0.025,
               mod_of_vector - height,
               16,
               1);

  glPopMatrix ();

  /* Draw head  */

  top_radius = height *
    tan (vector_arrow_angle * DEG2RAD);

  glPushMatrix ();

  glTranslatef (atoms[i].x + (mod_of_vector - height) * vector[0],
                atoms[i].y + (mod_of_vector - height) * vector[1],
                atoms[i].z + (mod_of_vector - height) * vector[2]);
  glRotatef (180.0, 0.0, 0.0, -1.0);
  glRotatef (angle, vector[1], -1.0 * vector[0], 0.0);

  gluCylinder (gl_quadric,
               top_radius,
               0,           /* base radius */
               height,
               16,          /* slices */
               8);          /* stacks */

  glPopMatrix ();

}


void gl_init (void)
{
  void bg_color_init(void); /* aro */

  XColor get_bg_color (void);

  double red, green, blue;

  Boolean color_initialized = False;

  static XColor bg_color_rgb;

  GLfloat light_position[] = {0.0, 0.0, 1.0, 0.0};

  if (color_initialized == False)
    {
      if(bg_color_parsed == 0 && strcmp(bg_color,"") != 0){
        bg_color_init();
        bg_color_parsed = 1;
      }

      bg_color_rgb = get_bg_color ();

      red   = bg_color_rgb.red   / 65535.0;
      green = bg_color_rgb.green / 65535.0;
      blue  = bg_color_rgb.blue  / 65535.0;
    }

  glClearColor (red, green, blue, 0.0);

  glShadeModel (GL_SMOOTH);

  /* Lighting */

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();

  glLightfv (GL_LIGHT0, GL_POSITION, light_position);
  glEnable (GL_LIGHT0);
  glDepthFunc (GL_LESS);

  glEnable (GL_COLOR_MATERIAL);
 
  /* Allocate a quadric */

  if (gl_quadric == NULL)
    {
      gl_quadric = gluNewQuadric ();
    }
}

void gl_open_file_init (void)
{
  void set_gl_eye(double);
  double get_gl_eye(void);
  double get_canvas_scale(void);
  Widget eye_scale;
  double new_gl_eye;
  
  /* 
     
     To keep same scale in gl as in x rendering, it appears that we
     should set gl_eye to be r_max * (canvas_width / 100.0) 

     In file.c:set_default_depth(void), we are setting 
        canvas_scale = canvas_width / (2.0 * (fudge = 2.)0 * r_max)
     => r_max = canvas_width / (canvas_scale * 2.0 * (fudge = 2.0))

     So we should have 
     gl_eye = (canvas_width)^2 / (canvas_scale * 2.0 * fudge * 100.0)

     note: When translating along z axis, there is a large difference
     between gl appearance and x appearance due to different depth
     factors.  Setting depth to 6 gives the x rendering roughly the
     same appearance as the gl rendering.
     
  */

  /* In following formula 400.0 is dependent on value of local var
     fudge in file.c:set_default_depth(void).  currently fudge = 2.0 */
  new_gl_eye = pow(canvas_width, 2.0) / (get_canvas_scale() * 400.0);
  
  set_gl_eye(new_gl_eye);

  if(gl_render_dialog){
    eye_scale = XtNameToWidget(gl_render_dialog, "row_column.eye_scale");
    XtVaSetValues(eye_scale, XmNvalue, (int) get_gl_eye(), NULL);
  }

}

void gl_set_perspective (void)
{

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();

  gl_update_near_and_far ();
  gluPerspective (gl_fov, 1.0, gl_perspective_near, gl_perspective_far);

}

void gl_update_near_and_far (void)
{
  double get_vector_scale (void);

  int i, j;
  
  double vector_scale = get_vector_scale ();
  double outline_offset = 0.1, shape_scale;
  Boolean max_set = False, min_set = False;

  /* Get the atoms with minimum/maximum Z for gluPerspective */

  for (i = 0; i < no_atoms; i++)
    {

      /* Take account of elliptical atoms */

      if (atoms[i].is_ellipse == 1)
        {
          shape_scale = atoms[i].max_shape;
        }
      else
        {
          shape_scale = 1.0;
        }

      /* Maximum distance of atom from origin; take in account the
         maximum radius of the atom since we don't want to clip it in
         the centre */

      if (max_set)
        {
          if ((atoms[i].z + (shape_scale * atoms[i].vdw_rad)) > gl_max_z)
            {
              gl_max_z = atoms[i].z + (shape_scale * atoms[i].vdw_rad);
            }
        }
      else
        {
          gl_max_z = atoms[i].z + (shape_scale * atoms[i].vdw_rad);
          max_set = True;
        }

      /* We use outline_offset since when in outline mode, the disk
         furthest from the eye disappears if some leeway isn't set */

      if (min_set)
        {
          if ((atoms[i].z - (shape_scale * atoms[i].vdw_rad)) < gl_min_z)
            {
              gl_min_z = (atoms[i].z - (shape_scale * atoms[i].vdw_rad)) - outline_offset;
            }
        }
      else
        {
          gl_min_z = (atoms[i].z - (shape_scale * atoms[i].vdw_rad)) - outline_offset;
          min_set = True;
        }
    }

  /* Now check the vectors */

  if (vector_flag == 1)
    {
      for (i = 0; i < no_atoms; i++)
        {
          for (j = 0; j < MAX_VECTORS_PER_ATOM; j++)
            {
              if (atoms[i].has_vector > j)
                {
                  if ((atoms[i].z + (atoms[i].v[j][2] * vector_scale)) > gl_max_z)
                    {
                      gl_max_z = atoms[i].z + (atoms[i].v[j][2] * vector_scale);
                    }
                  else if ((atoms[i].z + (atoms[i].v[j][2] * vector_scale)) < gl_min_z)
                    {
                      gl_min_z = atoms[i].z + (atoms[i].v[j][2] * vector_scale);
                    }
                }
            }
        }

      /* FIX ME -- we haven't taken into account the vector head,
         which may be clipped; fudge this */

      {
        double max_vector_size = 10.0;

        gl_max_z += max_vector_size;
        gl_min_z -= max_vector_size;
      }

    }

 
  if (gl_eye > gl_max_z)
    {
      gl_perspective_near = gl_eye - gl_max_z;
    }
  else
    {
      gl_perspective_near = 0.01; /* Small and positive */
    }

  if (gl_eye > gl_min_z)
    {
      gl_perspective_far = gl_eye - gl_min_z;
    }
  else
    {
      gl_perspective_far = 0.02; /* Small and slightly more positive */
    }

}

int get_gl_no_atom_segments (void)
{
  return (gl_no_atom_segments);
}

void set_gl_no_atom_segments (int new_value)
{
  gl_no_atom_segments = new_value;
}

int get_gl_no_bond_segments (void)
{
  return (gl_no_bond_segments);
}

void set_gl_no_bond_segments (int new_value)
{
  gl_no_bond_segments = new_value;
}

double get_gl_fov (void)
{
  return (gl_fov);
}

void set_gl_fov (double new_value)
{
  gl_fov = new_value;

  gl_set_perspective ();
}

double get_gl_eye (void)
{
  return (gl_eye);
}

void set_gl_eye (double new_value)
{
  gl_eye = new_value;

  gl_set_perspective ();
}

/* aro--> */

int get_gl_sep (void)
{
  return (gl_sep);
}

void set_gl_sep (int new_value)
{
  gl_sep = new_value;
}

/* Logical variable used to determine if GL rendering is active */

static Boolean render_using_gl = False;

void set_render_using_gl (Boolean value)
{
  render_using_gl = value;
}

Boolean render_using_gl_p (void)
{
  return (render_using_gl);
}

int get_gl_render_type (void)
{
  return (gl_render_type);
}

void set_gl_render_type (int new_value)
{
  
  gl_render_type = new_value;
    
}

/* aro--> */

/* Variable used to determine whether stereo rendering is active */
static enum render_stereo_types render_stereo_mode = NO_STEREO;

void set_gl_render_stereo(enum render_stereo_types value)
{
  render_stereo_mode = value;
}

enum render_stereo_types get_gl_render_stereo()
{
  return render_stereo_mode;
}
/* <--aro */

void gl_set_lighting (int new_value)
{
  gl_use_lighting = new_value;
}

int gl_get_lighting (void)
{
  return (gl_use_lighting);
}

void gl_set_specular_lighting (int new_value)
{
  GLfloat specular[] = {0.5, 0.5, 0.5, 1.0}; /* Configurable? */
  GLfloat no_specular[] = {0.0, 0.0, 0.0, 1.0};

  gl_use_specular_lighting = new_value;

  if (gl_use_specular_lighting)
    {
      glMaterialfv (GL_FRONT, GL_SPECULAR, specular);
      glMaterialfv (GL_FRONT, GL_SHININESS, gl_shininess);
    }
  else
    {
      glMaterialfv (GL_FRONT, GL_SPECULAR, no_specular);
      glMaterialfv (GL_FRONT, GL_SHININESS, gl_shininess);
    }
}

int gl_get_specular_lighting (void)
{
  return (gl_use_specular_lighting);
}


void gl_set_shininess (int new_value)
{
  GLfloat specular[] = {0.5, 0.5, 0.5, 1.0}; /* Configurable? */
  GLfloat no_specular[] = {0.0, 0.0, 0.0, 1.0};

  gl_shininess[0] = new_value;

  if (gl_use_specular_lighting)
    {
      glMaterialfv (GL_FRONT, GL_SPECULAR, specular);
      glMaterialfv (GL_FRONT, GL_SHININESS, gl_shininess);
    }
  else
    {
      glMaterialfv (GL_FRONT, GL_SPECULAR, no_specular);
      glMaterialfv (GL_FRONT, GL_SHININESS, gl_shininess);
    }
}

int gl_get_shininess (void)
{
  return (gl_shininess[0]);
}


#endif /* GL */
